@@ -75,7 +75,10 @@
             sweep_format_term/2,
             sweep_current_functors/2,
             sweep_term_search/2,
-            sweep_terms_at_point/2
+            sweep_terms_at_point/2,
+            sweep_predicate_dependencies/2,
+            sweep_async_goal/2,
+            sweep_interrupt_async_goal/2
           ]).
 
 :- use_module(library(pldoc)).
@@ -313,20 +316,39 @@ sweep_module_description([M0|P], [M|[P]]) :- atom_string(M0, M).
 sweep_predicate_references(MFN, Refs) :-
     term_string(M:F/N, MFN),
     pi_head(F/N, H),
-    findall([B,Path|Line],
+    findall([B, Path, From, Len],
             (xref_called(Path0, H, B0, _, Line),
              pi_head(B1, B0),
              term_string(B1, B),
-             atom_string(Path0, Path)),
+             atom_string(Path0, Path),
+             reference_span(Path0, Line, H, From, Len)),
             Refs,
             Tail),
-    findall([B,Path|Line],
+    findall([B, Path, From, Len],
             (xref_called(Path0, M:H, B0, _, Line),
              pi_head(B1, B0),
              term_string(B1, B),
-             atom_string(Path0, Path)),
+             atom_string(Path0, Path),
+             reference_span(Path0, Line, H, From, Len)),
             Tail).
 
+:- dynamic current_reference_span/2.
+
+reference_span(Path, Line, Head, From, Len) :-
+    retractall(current_reference_span(_, _)),
+    setup_call_cleanup(prolog_open_source(Path, Stream),
+                       (   prolog_source:seek_to_line(Stream, Line),
+                           prolog_colourise_term(Stream, Path, reference_span_(Head), [])
+                       ),
+                       prolog_close_source(Stream)),
+    !,
+    current_reference_span(From, Len).
+
+reference_span_(Head, goal_term(_, Goal), Beg0, Len) :-
+    \+ \+ Head = Goal,
+    Beg is Beg0 + 1,
+    assertz(current_reference_span(Beg, Len)).
+reference_span_(_, _, _, _) :- true.
 
 sweep_predicate_location(MFN, [Path|Line]) :-
     term_string(M:F/N, MFN),
@@ -358,6 +380,9 @@ sweep_predicate_apropos(Query0, Matches) :-
             Tail).
 
 sweep_predicate_location_(H, Path, Line) :-
+    xref_defined(Path0, H, How0),
+    xref_definition_line(How0, _),
+    xref_source(Path0, [comments(store)]),
     xref_defined(Path0, H, How),
     xref_definition_line(How, Line),
     !,
@@ -368,15 +393,26 @@ sweep_predicate_location_(H, Path, Line) :-
     atom_string(Path0, Path).
 
 sweep_predicate_location_(M, H, Path, Line) :-
-    (   xref_defined(Path0, M:H, How),
-        xref_definition_line(How, Line)
+    (   xref_defined(Path0, M:H, How0),
+        xref_definition_line(How0, _)
     ->  true
-    ;   xref_defined(Path0, H, How),
-        xref_definition_line(How, Line),
+    ;   xref_defined(Path0, H, How0),
+        xref_definition_line(How0, _),
         xref_module(Path0, M)
     ),
+    xref_source(Path0, [comments(store)]),
+    xref_defined(Path0, H, How),
+    xref_definition_line(How, Line),
     !,
     atom_string(Path0, Path).
+sweep_predicate_location_(M, H, P, L) :-
+    '$autoload':library_index(H, M, P0),
+    absolute_file_name(P0, P1, [extensions([pl])]),
+    xref_source(P1, [comments(store)]),
+    xref_defined(P1, H, How),
+    xref_definition_line(How, L),
+    !,
+    atom_string(P1, P).
 sweep_predicate_location_(M, H, Path, Line) :-
     predicate_property(M:H, file(Path0)),
     predicate_property(M:H, line_count(Line)),
@@ -470,14 +506,11 @@ sweep_color_normalized_(_, Goal0, [Kind0,Head|_], [Goal,Kind,F,N]) :-
     sweep_color_goal(Goal0),
     !,
     atom_string(Goal0, Goal),
-    sweeprolog_goal_kind_normalized(Kind0, Kind),
-    (   (   var(Head)
-        ->  true
-        ;   Head == []
-        )
-    ->  F = Head, N = 0
-    ;   pi_head(F0/N, Head),
+    sweep_goal_kind_normalized(Kind0, Kind),
+    (   callable(Head)
+    ->  pi_head(F0/N, Head),
         atom_string(F0, F)
+    ;   term_string(Head, F), N = 0
     ).
 sweep_color_normalized_(Offset, syntax_error, [Message0,Start0-End0|_], ["syntax_error", Message, Start, End]) :-
     !,
@@ -505,20 +538,59 @@ sweep_color_normalized_(_, file, [File0|_], ["file"|File]) :-
 sweep_color_normalized_(_, file_no_depend, [File0|_], ["file_no_depend"|File]) :-
     !,
     atom_string(File0, File).
+sweep_color_normalized_(_, type_error, [Kind0|_], ["type_error"|Kind]) :-
+    !,
+    Kind0 =.. [Kind1|_],
+    atom_string(Kind1, Kind).
 sweep_color_normalized_(_, Nom0, _, Nom) :-
     atom_string(Nom0, Nom).
 
-sweeprolog_goal_kind_normalized(autoload(Path0), ["autoload"|Path]) :-
+sweep_goal_kind_normalized(autoload(Path0), ["autoload"|Path]) :-
     !,
     absolute_file_name(Path0, Path1, [extensions([pl])]),
     atom_string(Path1, Path).
-sweeprolog_goal_kind_normalized(Kind0, Kind) :-
+sweep_goal_kind_normalized(imported(Path0), ["imported"|Path]) :-
+    !,
+    absolute_file_name(Path0, Path1, [extensions([pl])]),
+    atom_string(Path1, Path).
+sweep_goal_kind_normalized(global(Kind0, _), ["global"|Kind]) :-
+    !,
+    atom_string(Kind0, Kind).
+sweep_goal_kind_normalized(thread_local(_), "thread_local") :-
+    !.
+sweep_goal_kind_normalized(dynamic(_), "dynamic") :-
+    !.
+sweep_goal_kind_normalized(multifile(_), "multifile") :-
+    !.
+sweep_goal_kind_normalized(foreign(_), "foreign") :-
+    !.
+sweep_goal_kind_normalized(local(_), "local") :-
+    !.
+sweep_goal_kind_normalized(constraint(_), "constraint") :-
+    !.
+sweep_goal_kind_normalized(public(_), "public") :-
+    !.
+sweep_goal_kind_normalized(extern(Module0), ["extern",Module]) :-
+    !,
+    (   atom(Module0)
+    ->  atom_string(Module0, Module)
+    ;   Module = Module0
+    ).
+sweep_goal_kind_normalized(extern(Module0,Kind0), ["extern",Module,Kind]) :-
+    !,
+    (   atom(Module0)
+    ->  atom_string(Module0, Module)
+    ;   Module = Module0
+    ),
+    atom_string(Kind0, Kind).
+sweep_goal_kind_normalized(Kind0, Kind) :-
     term_string(Kind0, Kind).
 
 sweep_color_goal(goal).
 sweep_color_goal(goal_term).
 sweep_color_goal(head).
 sweep_color_goal(head_term).
+sweep_color_goal(predicate_indicator).
 
 sweep_expand_file_name([String|Dir], Exp) :-
     term_string(Spec, String, [syntax_errors(quiet)]),
@@ -773,8 +845,9 @@ sweep_beginning_of_next_predicate(Start, Next) :-
 
 sweep_source_id(Path) :-
     sweep_main_thread,
-    user:sweep_funcall("buffer-file-name", Path),
-    string(Path).
+    user:sweep_funcall("buffer-file-name", Path0),
+    string(Path0),
+    atom_string(Path, Path0).
 
 sweep_atom_collection(Sub, Col) :-
     findall(S,
@@ -900,15 +973,44 @@ sweep_string_to_atom(String, AtomString) :-
 sweep_file_path_in_library(Path, Spec) :-
     file_name_on_path(Path, Spec0),
     prolog_deps:segments(Spec0, Spec1),
-    term_string(Spec1, Spec).
+    (   string(Spec1)
+    ->  Spec = Spec1
+    ;   term_string(Spec1, Spec)
+    ).
 
+predicate_argument_names(M:F/A, Args) :-
+    (   M == system
+    ->  true
+    ;   sub_atom(M, 0, 1, _, '$')
+    ),
+    pldoc_man:load_man_object(F/A, _, _, DOM0),
+    memberchk(element(dt, _, DOM1), DOM0),
+    memberchk(element(a, _, DOM2), DOM1),
+    catch(findall(Arg,
+                  (   member(element(var, _, Vars), DOM2),
+                      member(ArgsSpec, Vars),
+                      term_string(CommaSeparatedArgs,
+                                  ArgsSpec,
+                                  [module(pldoc_modes),
+                                   variable_names(VN)]),
+                      maplist(call, VN),
+                      comma_list(CommaSeparatedArgs, ArgsList),
+                      member(Arg, ArgsList)
+                  ),
+                  Args0),
+          error(syntax_error(_),_),
+          fail),
+    predicate_argument_names_(A, Args0, Args).
 predicate_argument_names(M:F/A, Args) :-
     doc_comment(M:F/A, _, _, C),
     comment_modes(C, ModeAndDets),
     member(ModeAndDet, ModeAndDets),
     strip_det(ModeAndDet, Head),
     Head =.. [_|Args0],
-    length(Args0, A),
+    predicate_argument_names_(A, Args0, Args).
+
+predicate_argument_names_(Arity, Args0, Args) :-
+    length(Args0, Arity),
     maplist(strip_mode_and_type, Args0, Args).
 
 strip_mode_and_type(S, N), compound(S) => arg(1, S, N0), strip_type(N0, N).
@@ -1039,8 +1141,7 @@ list_tail([_|T0], T) :- nonvar(T0), T0 = [_|_], !, list_tail(T0, T).
 list_tail([_|T], T).
 
 sweep_terms_at_point([String, Start, Point], Res) :-
-    (   sweep_source_id(Path0),
-        atom_string(Path, Path0),
+    (   sweep_source_id(Path),
         findall(Op, xref_op(Path, Op), Ops),
         (   xref_module(Path, Module)
         ->  true
@@ -1088,3 +1189,43 @@ sweep_terms_at_point_(parentheses_term_position(_, _, SubPos), Start, Point, Beg
     sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
 sweep_terms_at_point_(quasi_quotation_position(_, _, _, SubPos, _), Start, Point, Beg, End) :-
     sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
+
+sweep_predicate_dependencies([To0|From0], Deps) :-
+    atom_string(To, To0),
+    atom_string(From, From0),
+    setof(PI,
+          PI0^Head^By^(
+                          xref_defined(To, Head, imported(From)),
+                          xref_called(To, Head, By),
+                          pi_head(PI0, Head),
+                          term_string(PI0, PI)
+                      ),
+          Deps).
+
+sweep_async_goal([GoalString|FD], TId) :-
+    term_string(Goal, GoalString),
+    random_between(1, 1024, Cookie),
+    thread_self(Self),
+    thread_create(sweep_start_async_goal(Self, Cookie, Goal, FD), T,
+                  [detached(true)]),
+    at_halt((   is_thread(T),
+                thread_property(T, status(running))
+            ->  thread_signal(T, thread_exit(0)),
+                thread_join(T, _)
+            ;   true
+            )),
+    thread_get_message(sweep_async_goal_started(Cookie)),
+    thread_property(T, id(TId)).
+
+sweep_start_async_goal(Caller, Cookie, Goal, FD) :-
+    thread_send_message(Caller, sweep_async_goal_started(Cookie)),
+    setup_call_cleanup((   sweep_fd_open(FD, Out),
+                           set_prolog_IO(current_input, Out, Out)
+                       ),
+                       once(Goal),
+                       (   format("~nSweep async goal finished~n"),
+                           close(Out)
+                       )).
+
+sweep_interrupt_async_goal(TId, TId) :-
+    thread_signal(TId, throw(interrupted)).
